//Team: 	Revolver 




/**
 * @file  
 * 	This file is the implementation of all the member functions of the 
 * 	FD class.
 * @brief
 * 	class FD function definitions
*/


//Include header file
#include "fd.h"

//************************* Member Functions **************************

/**
 * This function takes in a vector of search paths, and loops through
 * them, calling recursiveSearch each time, increasing encapsulation
 * of the recursiveSearch function
 * @param inSearchPaths a vector<fs::path>
*/
void FD::startSearch(vector<fs::path> inSearchPaths)
{

	/** @pre Vector of paths is not empty */
	/** @post recursiveSearch is called on all paths in vector */

	int searchSize = inSearchPaths.size();
	
	//ensure pre-condition
	preCondition((searchSize != 0), "FD::startSearch"); 

	//loop through all search paths
	for( unsigned int i = 0; i < searchSize; i++)
	{
		recursiveSearch(inSearchPaths[i]);
	}

	//Now check if any subdirectories match
	findDupDirs();	

}// end of startSearch function


/**
 * This function goes through fd_dirs and finds all dup dirs in a
 * top down manner. 
*/
void FD::findDupDirs()
{
	/** @pre Parent will always be before it's sub-direcotries in fd_dirs,
				Subdirectories will be contiguous, every fd_dir has a valid hash */
	/** @post Any duplicate directories will be stored in the first occurence
				 of that directory's entry in fd_dirs */

	//Variable(s)
	int matchRound = 0;			//Keeps track of which round of matching

	//Check if any subdirectories match, if not in files mode
	if ((fd_arguments & FILES_F) == 0)
   {
		int fdDirsSize = fd_dirs.size();

      //go through all directories and subdirectories
      for (unsigned int i = 0; i < fdDirsSize; i++)
      {
			//Check if we already found this directory's duplicates
			//by finding all dups of a duplicate directory earlier
			if (fd_dirs[i].foundMyDups == true)
				continue;			//if so, skip this one

			//If a child, get parent's matchRounds
			if (fd_dirs[i].parent != -1)
				fd_dirs[i].mRounds = fd_dirs[fd_dirs[i].parent].mRounds;
 
			//Go through all directories, and subdirectories of others
			//Notice that you start at i, and skip forward to next
			//parent. This way, you search every dir, and only once.
			for (unsigned int s = i; s < fdDirsSize; s++)
			{
				//if we are checking against self, skip ahead
				//past yourself, and all your sub directories.
				//Also, if my match round match any of s's, skip over.
				if ( s == i || hasMyMatch(fd_dirs[i].mRounds,fd_dirs[s].mRounds))
					s += fd_dirs[s].totalSubDirs;
				else
				{
					//Check if this directory matches.  
					//If not, move on.
					if (fd_dirs[i].hash == fd_dirs[s].hash)
					{
						//mark both dirs with this match round, and increment
						fd_dirs[i].mRounds.push_back(matchRound);
						fd_dirs[s].mRounds.push_back(matchRound);	
						matchRound++;						
					
						//Push directory s on directory i's vector
						fd_dirs[i].dupDirs.push_back(s);
					
						//Mark s as having had its dups found
						fd_dirs[s].foundMyDups = true;

						//Now since matched, skip over subs for s
						s += fd_dirs[s].totalSubDirs;
					}
				}
			}
		}
	}
}	 //end findDupDirs function


/**
 * This function goes through fd_files and prints out all files
 * which have duplicates, and what those duplicates are, in a nice
 * human-readable format
*/
void FD::printDups()
{

	/** @pre findDupDirs has already run */
	/** @post all duplicate directories are printed out together */

	int fdFilesSize = fd_files.size();

	//if there was an output file
	if(fd_arguments & OUTPUT_F)
	{
		fs::fstream outStream(fd_cmdArgs.outputFile, ios::out | ios::app);
		
		for (unsigned int i = 0; i < fdFilesSize; i++)
		{
			if ( !fd_files[i].duplicates.empty() && shouldPrint(i))
			{
				outStream << "\nFile" << endl;
				outStream << "  " << fd_files[i].path.native_file_string() << endl;
				outStream << "has duplicate file(s)..." << endl;

     		   for (unsigned int j = 0; j < fd_files[i].duplicates.size(); j++)
				{
					outStream << "  " 
								 << fd_files[i].duplicates[j].path.native_file_string()
					  	  	    << endl;
				}
			}//end of if it has duplicates

		}//end of outside for

	}//end of if we have an output file

	
	// else we're printing to stdout
	else
	{
		for (unsigned int i = 0; i < fdFilesSize; i++)
		{

			if ( !fd_files[i].duplicates.empty() && shouldPrint(i))
			{
				cout << "\nFile" << endl; 
				cout << "  " << fd_files[i].path.native_file_string() << endl;
				cout << "has duplicate file(s)..." << endl;

     	   	for (unsigned int j = 0; j < fd_files[i].duplicates.size(); j++)
				{
					cout << "  " 
						  << fd_files[i].duplicates[j].path.native_file_string()
					  	  << endl;
				}
			}//end of if it has duplicates

		}//end of outside for
			
	}//end of else it's stdout

}// end of printDups function


/**
 * This function takes in an integer index into fd_files and 
 * determines whether it should be printed or not, based on
 * the directories each file is included in.  It returns true
 * if the file index should be printed out, false otherwise
 * @param inFileIndex an int
 * @return bool
*/
bool FD::shouldPrint(int inFileIndex)
{

	/** @pre inFileIndex is a valid index into fd_files */
	/** @post returns true if an instance of that file exists outside
		 		 a group of duplicate directories */


	int fdDirsSize = fd_dirs.size();

	//if the --files flag is set, always print
	if((fd_arguments & FILES_F) != 0)
		return true;

	//flag for if it's in at least one dupDir
	bool inDupDir = false;

	//loop through all directories	
	for(unsigned int i = 0; i < fdDirsSize; i++)
	{
		//only check against directories with dups
		if(fd_dirs[i].dupDirs.size() == 0)
			continue;

		//now we're at a dup directory entry, store all indeces for this
		//group of dirs
		vector<int>dirGroup = fd_dirs[i].dupDirs;
		dirGroup.push_back(i);
		
		//only worry about it if one of the directories in
		//dirGroup has inFileIndex
		if(!isIntInVect(inFileIndex, fd_dirs[i].fGroups))
			continue;

		//if we get here, the file is in a dupDir
		inDupDir = true;

		//now see if any files in this group are not found in
		//dirGroup, if so, need to print it out
		if(!isDirInVect(fd_files[inFileIndex].dirIndex, dirGroup))
			return true;
		
		//search duplicates
		vector<FileDup> dups = fd_files[inFileIndex].duplicates;

		int dupsSize = dups.size();

		for(unsigned int j = 0; j < dupsSize; j++)
		{

			//if any of them are not in dirGroup, need to print
			if(!isDirInVect(dups[j].dirIndex, dirGroup))
				return true;

		}//end of for all duplicates

	}//end of for all directories
	
	//if we got here, don't have to print it
	return !inDupDir;	
	
}//end of shouldPrint function


/**
 * This function goes through fd_dirs and prints out all dirs
 * which have duplicates, and what those duplicates are, in a nice
 * human-readable format
*/
void FD::printDupDirs()
{
	/** @pre TRUE */
	/** @post all duplicate directories were printed out to either stdout
				 or appropriate output file, based on fd_arguments */

	int fdDirsSize = fd_dirs.size();

	//if there was an output file
   if(fd_arguments & OUTPUT_F)
   {	
		//make output stream
		fs::fstream outStream(fd_cmdArgs.outputFile);

		//Print out dirs with dup dirs
		for (unsigned int y = 0; y < fdDirsSize; y++)
		{
			//If it has dup dirs, print out
			if (fd_dirs[y].dupDirs.size() != 0)
			{
				outStream << "\nDirectory" << endl;
   			outStream << "  " << fd_dirs[y].dirPath.native_file_string();
      		outStream << "\nHas duplicate directory(ies)..." << endl;
      
				for (unsigned int p = 0; p < fd_dirs[y].dupDirs.size(); p++)
				{
					fs::path tempPath = fd_dirs[fd_dirs[y].dupDirs[p]].dirPath;
      			outStream << "  " << tempPath.native_file_string() << endl;
				}
   		}
		}
		
		//close outStream
		outStream.close();

	}
	else		//No output file, so print to std out
	{
		//Print out dirs with dup dirs
      for (unsigned int y = 0; y < fdDirsSize; y++)
      {
         //If it has dup dirs, print out
         if (fd_dirs[y].dupDirs.size() != 0)
         {
            cout << "\nDirectory" << endl;
            cout << "  " << fd_dirs[y].dirPath.native_file_string() << endl;
            cout << "Has duplicate directory(ies)..." << endl;

            for (unsigned int p = 0; p < fd_dirs[y].dupDirs.size(); p++)
				{
					fs::path tempPath = fd_dirs[fd_dirs[y].dupDirs[p]].dirPath;
               cout << "  " << tempPath.native_file_string() << endl;
				}
         }
      }
	}	

} //end of printDupDirs function


/**
 * This function parses the command line and populates the vector of paths
 * @param inArgc a int
 * @param inArgv a char* [] 
*/
void FD::parseCommandLine( int inArgc, char* inArgv[])
{

	/** @pre TRUE */
	/** @post All command line agruments were stored in fd_arguments,
				 or program termintaed because of bad command line arguments */

	//temp variable to know if the flag was found
	bool flagFound = false;
	if(inArgc > 1)
	{

		//loop through all args, leaving out executable
		for(int i = 1; i < inArgc; i++)
		{
			//Copy the element into a string for easier manip
			string temp = inArgv[i];
			if(temp[0] == '-')				//We are at state 1 "-" type param
			{
				if(temp[1] == '-')			//We are at state 2 "--" type param
					flagFound = isFlag(temp,2);		//Check for long params
				else
					flagFound = isFlag(temp,1);		//Check for small params

				if(flagFound)
				{
					//Check if the help command has been called
					if(fd_arguments & HELP_F)
					{
						printHelp();		//This will terminate the application 
					}
					else if(fd_arguments & VERSION_F)
					{
						printVersion();	//This will terminate the application
					}
					else if(fd_arguments & INPUT_F)
					{
						getInput(i, inArgc, inArgv);
						fd_arguments &= ~INPUT_F;	//remove this flag
					}
					else if(fd_arguments & OUTPUT_F)
					{
						getOutput(i, inArgc, inArgv);
					}
				}
				else									//We got an invalid parameter
				{
					cout << "Invalid parameter!\nUse fd --help for options\n";		
					exit(0);
				}
			}//end of if looking for parameters
			else//If we get here, we must only have paths left
			{
				//store in a temp variable
				string tempStr = inArgv[i];	
	
				//expand the symbols
				expandSymbols(tempStr);
				
				//Add the paths to paths member of fd_cmdArgs
				fs::path tempPath(tempStr);

				//only add valid paths that don't already exist
				if(fs::exists(tempPath))
					fd_cmdArgs.paths.push_back(tempPath);

				else
					cout << "no such path: " << tempPath.string() << endl;				
			}
		}//end of for loop

	}//end of if we have arguments
	
}//end of parseCommandLine function


/**
 * This function prints out the help file telling the user how to
 * use the fd command line options 
*/
void FD::printHelp()
{
	/** @pre TRUE */
	/** @post help info was printed to STDOUT */

	//The actual call to the help file would go here
	cout	<< "Usage: fd [options] [paths]\n"
			<< "Options:\n"
			<< "\t-i (--input)\tSpecify input file\n"
			<< "\t-o (--output)\tSpecify output file\n"
			<< "\t-v (--version)\tGive current version of fd\n"
			<< "\t-h (--help)\tDisplay this help file\n"
			<< "\t-p (--paranoid)\tUse byte-by-byte comparison instead of hashing"
			<< "\n\t-f (--files)\tExpand duplicate sub-trees to show all files"
			<< endl;

	exit(0);
}


/**
 * This function prints the version of the current fd
*/
void FD::printVersion()
{
	/** @pre TRUE */
	/** @post Version info was printed to STDOUT */

	//This is a temp variable, later on the program will know its own
	//Version
	cout << "FD VERSION " << VERSION << "\n";
	cout << "Written By Team Revolver\n";
	cout << "This is free software.\n";
	exit(0);
}


//Following code taken from Boost.org, with modifications
/**
 * This function takes in a path and index, and iterates through the directory.
 * If it encounters a subdirectory, it recursively searches that one,
 * hence the name. Pushes info onto global directory tree.
 * @param inDirPath an fs::path 
 * @return int
*/
int FD::recursiveSearch(const fs::path &inDirPath)
{

	/** @pre inDirPath is a valid path */
	/** @post inDirPath and all sub-directries have been procesed */

	//This will be index of this directory
	int myIndex = 0;

  	//see if current path is a directory, 1st instance is dir
  	if( fs::is_directory (inDirPath) )
  	{	
		//Make a new directory object
		FD::Directory thisDir(inDirPath);

		//Push this direcory object to global directory list
		fd_dirs.push_back(thisDir);

		//Get updated index for this directory
		myIndex = fd_dirs.size() - 1;

   	//create iterator, without path creates path-to-end iterator
      fs::directory_iterator endIter;
      for( fs::directory_iterator iter (inDirPath); iter != endIter; ++iter )
      {
     		//check if this path is a directory
         if(fs::is_directory( *iter ))
			{		
				//Call recursiveSearch which returns child's index
         	int childInd = recursiveSearch(*iter);

				//Store all of child's file groups in this dir's file groups
				for(unsigned int x = 0; x < fd_dirs[childInd].fGroups.size(); x++)
					fd_dirs[myIndex].fGroups.push_back(fd_dirs[childInd].fGroups[x]);

				//Put this child directory in list of subdirectories
				fd_dirs[myIndex].subDirs.push_back(childInd);

				//Push this index as parent of child directory
				fd_dirs[childInd].parent = myIndex;

				//Update total number of sub directories, counting all child's
				//subdirectories. Then add one for child directory.
				fd_dirs[myIndex].totalSubDirs += fd_dirs[childInd].totalSubDirs+1;
			}
         else   
         {
         	//Check if this file is already present in list. 
				//If so, is a duplicate
				bool results = false;
            bool match = false;
				int index = 0;

				string iterHash = "";

				//Calculate hash for this iterator, if not in paranoid mode
				if ((fd_arguments & PARANOID_F) == 0)
					 iterHash = getHash(*iter);

				int fdFilesSize = fd_files.size();

				//comparison code
            for(unsigned int i = 0; i < fdFilesSize && match == false; i++)
				{
					//Check if in paranoid mode. If so, use byte-by-byte
					if ((fd_arguments & PARANOID_F) != 0)
            		results = areFilesSame(*iter, fd_files[i].path); 
					else
						results = areFilesSameHash(iterHash, fd_files[i].hash);

					if (results)	//If files are the same
					{
						match = true;
						index = i;	
					}
				}

            if (match)      				//If file is a duplicate
            {
            	//push this new path and index onto the file's 
					//list of duplicates.
					FD::FileDup dupTemp(*iter, myIndex);
               fd_files[index].duplicates.push_back(dupTemp);

					//Add this group to directory's fGroups vector
					fd_dirs[myIndex].fGroups.push_back(index);
				}
            else
            {
            	//Is a unique file, so add to list of fd_files
					FD::File temp(*iter, iterHash, myIndex);
 					fd_files.push_back(temp);

					//Add this group to directory's fGroups vector
               fd_dirs[myIndex].fGroups.push_back(fd_files.size() - 1);
            }     
				//Notice: This function does NOT check for dup dirs.
			}
		}
   }
	
	//Now that we have all the file groups for this dir, calculate hash.
	fd_dirs[myIndex].hash = getDirHash(fd_dirs[myIndex].fGroups);

	//Return this directory's index	
	return myIndex;		

}//end of recursiveSearch function


/**
 * This function searches through library of proper parameters 
 * and decides if it is a proper parameter, updates the arguments
 * variable.
 * @param inStringToTest a string.
 * @param type_of_param a integer.  
*/
bool FD::isFlag(string inStringToTest, int inTypeOfParam)
{

	/** @pre inTypeOfParam is 1 or 2 */
	/** @post Returns true if inStringToTest is a flag of type
		 inTypeOfParam, either short or long */

	//ensure pre-condition
	preCondition((inTypeOfParam == 1 || inTypeOfParam == 2), "FD::isFlag");

	//current number of parameters
	vector<string> parameters;
	parameters.push_back("help");
	parameters.push_back("version");
	parameters.push_back("input");
	parameters.push_back("output");
	parameters.push_back("paranoid");
	parameters.push_back("files");
	const int numOfParams = (int)parameters.size();

	//String used of comparisons
	string tempString;

	//Decide what type of parameter it is
	//if it's a short one
	if(inTypeOfParam == 1)
	{
		//Check to see that there is no more than one letter 
		//after the '-'
		if(inStringToTest.size() != 2)
		{
			cout << "Invalid parameter!\nUse fd --help for options\n";
			exit(0);
		}
		else
			tempString = inStringToTest[1];
	}

	//else it's a long param
	else if(inTypeOfParam == 2)
	{
		tempString = inStringToTest.substr(2);
	}
	
	//if wrong typeOfParam is passed, exit will never be reached b/c of
	//pre-condition
/*	else 	
	{												
		cout << "An Invalid argument was passed to is flag.\n";
		exit(0);											
	}
*/
	//Search through my array of parameters and find it matches any of them
	//setting that flag at the same time
	for(int i = 0; i < numOfParams; i++)
	{
		string::size_type pos = parameters[i].find(tempString,0);

		//If it is a proper flag
		if(pos == 0)
		{
		
			//check to make sure that the sub-string is unique
			if( !isUniqueSubStr(tempString, parameters, i) )
			{
				cout << "Sub-string matches multiple commands : " 
					  << tempString << endl;
				exit(0);
			}	

			//Set the proper flag
			switch(i)
			{
				case 0:
					fd_arguments |= HELP_F;
					break;
				case 1:
					fd_arguments |= VERSION_F;
					break;
				case 2:
					fd_arguments |= INPUT_F;
					break;	
				case 3:
					//if we already had an output file then report an error
					//and terminate
					if(fd_arguments & OUTPUT_F)
					{
						cout << "Multiple output files are not permitted!\n";
						exit(0);
					}
					fd_arguments |= OUTPUT_F;
					break;
				case 4:
					fd_arguments |= PARANOID_F;
					break;
				case 5:
					fd_arguments |= FILES_F;
					break;
				default:		
					//This should never happen
					cout << "There was an error comparing parameters!\n";
			}
			//if we get here, then it was a valid flag
			return true;

		}//end of if it is this valid arg

	}//end of for all valid args

	//If we get here then there was an invalid parameter
	return false;

}//end isFlag function


/**
 * This function gets the path for the input file and checks if it 
 * exists. Adding what paths it contains to fd_cmdArgs.paths
 * @param inOutCurrentCounter an int
 * @param inArgc an int
 * @param inArgv a char * []
*/
void FD::getInput(int &inOutCurrentCounter, int inArgc, char* inArgv[])
{

	/** @pre -i or --input flags were found in inArgv at inOutCurrentCounter */
	/** @post all input paths from file entered by user have been saved */

	//Look at the next member of inArgv
	//Must not be empty
	inOutCurrentCounter++;

	//Must use a try-catch sequence here
	bool hasFileName = ( inOutCurrentCounter <= (inArgc - 1) );
	bool hasValidFileName = false;
	if(hasFileName)
	{
		string fileNameStr = inArgv[inOutCurrentCounter];
		Cfunc cUtility;
		int flags = cUtility.read;
		if(cUtility.canOpenFile(fileNameStr, flags))
		{
			//signal that we did get a valid filename
			hasValidFileName = true;

			//expand ./ etc
			expandSymbols(fileNameStr);
	
			//create a path object	
			fs::path inputFilePath(fileNameStr);
		
			//store the paths from it into fd_cmdArgs.paths
			getPathsFromFile(inputFilePath);
			
		}

	}//end of if it has a filename

	//if a valid filename was not given	
	if( (!hasFileName) || (!hasValidFileName) )
	{
		cout << "Option -i must be followed by a valid name "
			  << "of a readable file!\n";
		exit(0);
	}

}//end of getInput function


/**
 * This function gets the path for the output file,
 * it will also check wether or not another output file had been given.
*/
void FD::getOutput(int &inOutCurrentCounter, int inArgc, char* inArgv[])
{

	/** @pre -o or --output flags were found in inArgv at position
				inOutCurrentCounter */
	/** @post if output file existed, a backup is made, otherwise
				 the path to the output file is saved in fd_cmdArgs.outFile */
				 
	//Look at the next member of argv
	//Must not be empty
	inOutCurrentCounter++;

	bool hasFileName = ( inOutCurrentCounter <= (inArgc - 1) );
	bool hasValidFileName = false;
	if( hasFileName )
	{
		string fileName = inArgv[inOutCurrentCounter];
		Cfunc cUtility;
		int flags = cUtility.readWrite | cUtility.create | cUtility.append;
		if(cUtility.canOpenFile(fileName, flags))
		{
			hasValidFileName = true;

			//expand ./ etc
			expandSymbols(fileName);

			fs::path outputFilePath(fileName);
			fs::ifstream tempStream(outputFilePath);

			//if there is anything in the file, copy it to backup
			//and create a new file with that name
			string temp = "";			
			if(tempStream >> temp)
			{
				cUtility.createBackup(fileName);

			}//end of if output file has content			

			fd_cmdArgs.outputFile = outputFilePath;
			
		}//end of if file can be opened	
	
	}//end of if filename is there

	//if we didn't get a valid filename, say so and exit
	if( (!hasFileName) || (!hasValidFileName) )
	{
		cout << "Option -o must be followed by a valid name"
			  << " of a writable file!\n";
		exit(0);
	}

}//end of getOutput function


/**
 * This function takes in an fs::ifstream object, and reads in all
 * the paths from it, storing them in fd_cmdArgs.paths
 * @param inInputFile an fs::path
*/
void FD::getPathsFromFile(fs::path inInputFile)
{

	/** @pre inInputFile is the path of an input file containing paths */
	/** @post All paths in inInputFile have been stored */

	//create an fs::ifstream
	fs::ifstream input(inInputFile);	

	string tempStr = "";
	while(input >> tempStr)
	{
		//expand the symbols
		expandSymbols(tempStr);
	
		//push back each path into fd_cmdArgs.paths
		fs::path tempPath(tempStr);

		//only add valid paths that are no already in 
		if(fs::exists(tempPath))
			fd_cmdArgs.paths.push_back(tempPath);		

		else
			cout << "no such path: " << tempPath.string() << endl;

	}//end of for all input

}//end of getPathsFromFile function


/**
 * This funciton takes in a dir index and a vector of
 * dir indeces, and checks if the first index, or any of
 * its parents are in the vector, returning true if it is,
 * false otherwise
 * @param inDirIndex an int
 * @param inDirs a vector<int>
 * @return bool
*/
bool FD::isDirInVect(int inDirIndex, vector<int> inDirs)
{

	/** @pre inDirIndex is a valid index into fd_dirs*/
	/** @post Returns true if inDirIndex or any of it's parents are
				 in inDirs, otherwise returns false */		

	//loop through fist vector
	for(unsigned int i = 0; i < inDirs.size(); i++)
	{
			
		//go through all parents of inDirIndex
		int currentDir = inDirIndex;
		while(currentDir != -1)
		{
			//if it matched, return true
			if(currentDir == inDirs[i])
				return true;
			
			//move to your parent
			currentDir = fd_dirs[currentDir].parent;
			
		}//end of while there are parents
		

	}//end of for first vector

	//if we get here, no match
	return false;
	
}//end of isDirInVect function


//***********************************************************************


//*********************** Non-Member Functions **************************

/**
 * This function expands '~' to the apropriate home directory
 * as well as expanding '..' to the parent directory
 * @param inOutPathName a string
*/
void expandSymbols(string &inOutPathName)
{

	/** @pre inOutPathName is a valid path name with or without wildcards */
	/** @post inOutPathName is a valid path name with all wildcards expanded */

	//instance of Cfunc class
	Cfunc cUtility;
	
	//check for "~"
	string::size_type pos = inOutPathName.find("~", 0);
	if(pos != string::npos)
		inOutPathName.replace(pos, 1, cUtility.getHome());

	//check for '../'
	pos = inOutPathName.find("../", 0);
	if(pos == 0)
	{
		string parentPath = fs::current_path().string() + "/../";
		inOutPathName.replace(pos, 3, parentPath);
	}

	//check for './'
	pos = inOutPathName.find("./", 0);
	if(pos == 0)
		inOutPathName.replace(pos, 2, fs::current_path().string() + "/");

	//check for '.'
	if(inOutPathName == ".")
		inOutPathName = fs::current_path().string();


	//if at this point, the string doesn't start with '/', add
	//current directory
	if(inOutPathName[0] != '/')
		inOutPathName = fs::current_path().string() + "/" + inOutPathName;

}//end of expandSymbols function


/**
 * This function takes in two paths, representing a reference file
 * and a test file, compares them and returns true if they are the same,
 * false if they are different, based on the criteria of size and contents
 * @param inRefPath an fs::path
 * @param inTestPath an fs::path
 * @return bool
*/
bool areFilesSame(fs::path inRefPath, fs::path inTestPath)
{	

	/** @pre inRefPath and inTestPath are two valid file paths */
	/** @post Returns true if files have same content, false otherwise */

	//first check if they are the same size, if not, return false
	Cfunc cUtility;
	if( !cUtility.areFilesSameSize(inRefPath.string(), inTestPath.string()) )
		return false;

	//local constants for is or isn't duplicate file
	const bool lkIsDuplicate = true;
	const bool lkIsNotDuplicate = false;
	
	//first make fstream objects
	fs::ifstream refFile(inRefPath);
	fs::ifstream testFile(inTestPath);

	//now go through comparing one byte at a time
	char refByte = 'a', testByte = 'a';
	while(refFile.get(refByte))
	{

		//read in testByte
		testFile.get(testByte);

		//now compare bytes
		if(testByte != refByte)
			return lkIsNotDuplicate;

	}//end of for whole refFile

	//if we get to here, then it is a duplicate file
	return lkIsDuplicate;

} //end of areFilesSame function


/**
* This function takes in a path, calculates a hash from it, and returns it
* as a hexadecimal string.
* @param filePath is an fs::path
* @return string
*/
string getHash(fs::path filePath)
{
	
	/** @pre filePath is a valid path to a file */
	/** @post Returns an md5 has for file at filePath */

	//Create ifstream object from path
   ifstream file1(filePath.native_file_string().c_str());

   //Calculate hash for file
   MD5 h1(file1);
 
   //Store hash in string
   string hash = h1.hex_digest();

	//close file
	file1.close();	

	//return string
	return hash;

} 	//end getHash function


/**
* This function takes in a directory's vector of files groups,
* sorts this vector, calculates a hash from it, and then returns it
* as a hexadecimal string.
* @param inVect is a vector of type int
* @return string
*/
string getDirHash(vector<int> inVect)
{
	/** @pre TRUE */
	/** @post Returns a unique md5 hash of ints in inVect, where order doesn't
				 matter */

	//first sort the vector
	sort(inVect.begin(), inVect.end());

	//Cfunc object for C-Stuff
	Cfunc cUtility;
	cUtility.createTempFile();

	//declare ofstream object
	ofstream tempFileOut("fd.temp");

	for(unsigned int i = 0; i < inVect.size(); i++)
		tempFileOut << inVect[i] << ",";

	tempFileOut.close();

	//declare ifstream object
	ifstream tempFileIn("fd.temp");

	//now get a hash
	MD5 h1(tempFileIn);

	string hash = h1.hex_digest();

	tempFileOut.close();

	//get rid of the temp file
	cUtility.deleteTempFile();

	return hash;	

}//end of getDirHash function


/**
* This function takes in two string, hashes, and compares them.
* Returns true if hashes are the same, false otherwise.
* @param iterHash is a string
* @param vectorHash is a string
* @return bool
*/
bool areFilesSameHash(string iterHash, string vectorHash)
{
	/** @pre file in vector has a non-empty hash */
	/** @post if files were the same, it will return true */
	
	//ensure pre-condition
	preCondition((vectorHash != ""), "areFilesSameHash");
	
	//Compare both hashes, and return comparison result
	return iterHash == vectorHash;

}//end areFilesSameHash function


/**
 * This function takes in a string which is the substring in question,
 * a vector of accepted parameter strings, and the index into that vector
 * which the parameter matched.  It then makes sure that no other parameters
 * also match that substring
 * @param inSubStr a string
 * @param inValidArgs a vector<string>
 * @param inIndex an int
 * @return bool
*/
bool isUniqueSubStr(string inSubStr, vector<string> inValidArgs, int inIndex)
{

	/** @pre inIndex is the location of the valid argument that inSubStr
				first matched, and inValidArgs is a vector of all valid argument
				strings */
	/** @oost returns true if there are no other valid argument strings that
				 match inSubStr, false otherwise */

	//make sure inIndex is in range
	string func = "isUniqueSubStr";
	bool condition = (0 <= inIndex && inIndex <= (int)inValidArgs.size());
	preCondition(condition, func);

	//go through rest of vector, checking for inSubStr
	for(unsigned int i = (unsigned int)inIndex + 1; i < inValidArgs.size(); i++)
	{

		//if pos is 0, then not unique
		string::size_type pos = inValidArgs[i].find(inSubStr,0);
		if(pos == 0)
			return false;

	}//end of for rest of valid args

	//if we get here, it is unique
	return true;

}//end of isUniqeSubStr function


/**
 * This function takes in two vectors and checks if they 
 * have any common members. Should be O(n) time.
 * @param v1 a vector<int>
 * @param v2 a vector<int>
 * @return bool
*/
bool hasMyMatch(vector<int> v1, vector<int> v2)
{
	/** @pre TRUE */
	/** @post returns true if v1 and v2 have a number in common, 
				 false otherwise */

	//Variables
	vector<bool> u;
	int uSize = 0;
	int v1Size = v1.size();
	int v2Size = v2.size();

	//Check if either vector is empty
	if ( v1Size == 0 || v2Size == 0)
		return false;

	//Go through v1 and put all its members into proper slot
	for (int i = 0; i < v1Size; i++)
	{
		//loop until you find space to mark v1[i]
		while (uSize <= v1[i])
		{
			u.push_back(0);
			uSize++;
		}

		//Mark this slot as holding a value
		u[v1[i]] = true;
	}

	//Now, for every member of v2 check proper slot
	for (int j = 0; j < v2Size; j++)
	{
		//Don't step off of array
		if ( v2[j] >= uSize)
			continue;

		//Check if this spot already marked
		if (u[v2[j]] == true)
				return true;
	}

	return false;

}	//end hasMyMatch


/**
 * This function takes in an integer to check for and a vector
 * to check in, then checks for the integer in the vector, 
 * returning true if it finds it, false otherwise
 * @param inInt an int
 * @param inVect a vector<int>
 * @return bool
*/
bool isIntInVect(int inInt, vector<int> inVect)
{

	/** @pre TRUE */
	/** @post Returns true if inInt is in inVect, false otherwise */

//	int inVectSize = inVect.size();

	vector<int>::iterator vend = inVect.end();

	//loop through vector
//	for(unsigned int i = 0; i < inVectSize; i++)
	for (vector<int>::iterator iter = inVect.begin(); iter != vend; iter++)
	{
//		if(inVect[i] == inInt)
		if(*iter == inInt)
			return true;

	}//end of for inVect

	return false;
	
}//end of isIntInVect function


/**
 * This function takes in a fs::path and checks if it is in the
 * given vector of paths, returning true if it is, false if it's not
 * @param inPath an fs::path
 * @param inVect a vector<fs::path>
 * @return bool
*/
bool isPathInVect(fs::path inPath, vector<fs::path> inVect)
{
	/** @pre TRUE */
	/** @post if inPath was in inVect returned true, else false */

	for(unsigned int i = 0; i < inVect.size(); i++)
	{
		if(inPath.string() == inVect[i].string())
			return true;
	}

	//if you get here, false
	return false;

}//end of isPathInVect function


/**
 * This function goes through the vector of strings, and removes all
 * those reperesenting sub-directories of already existing paths in
 * the vector.  Also removing repeated paths.
 * @param inOutPaths a vector<fs::path>
*/
void removeRedundantPaths(vector<fs::path> &inOutPaths)
{
	/** @pre TRUE */
	/** @post There are no repeated paths or sub-dirs in vector*/
	
	//hold size
	const unsigned int size = inOutPaths.size();

	//decide which to keep
	bool isRedundant[size];
	for(unsigned int i = 0; i < size; i++)
		isRedundant[i] = false;	

	for(unsigned int i = 0; i < size; i++)
	{
		//hold current path
		string refPath = inOutPaths[i].string();
		
		for(unsigned int j = i+1; j < size; j++)
		{
			//hold testPath
			string testPath = inOutPaths[j].string();
		
			if(testPath == refPath)
				isRedundant[j] = true;
	
			//find any subStrs
			string::size_type pos1 = refPath.find(testPath + "/", 0);
			string::size_type pos2 = testPath.find(refPath + "/", 0);

			if(0 == pos1)
				isRedundant[i] = true;
			else if(0 == pos2)
				isRedundant[j] = true;

		}//end of inside loop
	
	}//end of outside loop

	//now store paths and clear inOutPaths, then refill with uniques
	vector<fs::path> tempPaths = inOutPaths;
	inOutPaths.clear();

	for(unsigned int i = 0; i < size; i++)
	{
		if(!isRedundant[i])
			inOutPaths.push_back(tempPaths[i]);
	}

}//end of removeRedundantPaths function
