//Team: 	Revolver 




/**
 * @file
 * 	This file is the header for our FD class which handles most of the
 * 	functionality of the fd utility
 * @brief
 * 	header for FD class
*/


//*********************** Includes and namespace *************************
#include "/usr/local/pkgs/boost-1.30.2/boost/filesystem/operations.hpp"
#include "/usr/local/pkgs/boost-1.30.2/boost/filesystem/fstream.hpp"
#include <vector>
#include <string>
#include <iostream>
#include <algorithm>
#include <cstdio>
#include "cFunc.h"
#include "md5.h"
#include "condition.h"

/**
 * @brief boost::filesystem namespace
 * @namespace fs
 * This namespace is the boost::namespace used for objects
 * and functions included from the boost filesystem library
*/
namespace fs = boost::filesystem;
using namespace std;

//************************************************************************

//if not defined, define
#ifndef _FD_H_
#define _FD_H_

//****************************** Constants *******************************
const int HELP_F = 1;
const int VERSION_F = 2;
const int INPUT_F = 4;
const int OUTPUT_F = 8;
const int PARANOID_F = 16;
const int FILES_F = 32;
const string VERSION = "2.2";

//************************************************************************


//**************************** Utility functions *************************

/// expands symbols like '~' and '..' and '.'
void expandSymbols(string &inOutPathName);

/// compares files byte to byte
bool areFilesSame(fs::path inRefFile, fs::path inTestFile);

/// compares files using MD5 hash function
bool areFilesSameHash(string iterHash, string vectorHash);

/// returns a hash from a given file path
string getHash(fs::path filePath);

// returns a hash from a given vector of ints
string getDirHash(vector<int> inVect);

/// tests if substring is also a substring of another valid arg
bool isUniqueSubStr(string inSubStr, vector<string> inValidArgs, int inIndex); 

/// takes in two vectores, and sees if they have any common members
bool hasMyMatch(vector<int> v1, vector<int> v2);

/// Takes in an int and checks if it's in vector
bool isIntInVect(int inInt, vector<int> inVect);

/// Takes in an fs::path  and checks if it's in vector
bool isPathInVect(fs::path inPath, vector<fs::path> inVect);

/// Takes in a vector of strings holding paths and removes redundancies
void removeRedundantPaths(vector<fs::path> &inOutPaths);

//************************************************************************


//*************************** Class Declaration **************************

/**
 * @class FD fd.h "fd.h"
 * @brief FD class, main implementation of fd
 * The FD class contains the main implementation of the fd utility
 * allowing encapulation of the gritty details
*/
class FD{
	public:

		/// FD constructor
		FD() { fd_arguments = 0; }

		/// FD destructor
		~FD() {}

		/// loops through all paths, calling recursiveSearch
		void startSearch(vector<fs::path> inSearchPaths);

		/// prints out list of files with their duplicates
		void printDups();

		/// tests whether these dups need to be printed
		bool shouldPrint(int inFileIndex);
		
		/// prints out list of directories with their dups
		void printDupDirs();

		/// parses the command line
		void parseCommandLine(int inArgc, char* inArgv []);

		/// returns fd_arguments
		int getArgs() { return fd_arguments; }

		/// returns a vector of paths from arguments
		vector<fs::path> getPathsFromArgs() { return fd_cmdArgs.paths; }

		/// returns true if there were command line paths
		bool areCmdLinePaths() { return (fd_cmdArgs.paths.size() != 0); }

		/// Checks if inDirIndex or it's parents are in inDirs
		bool isDirInVect(int inDirIndex, vector<int> inDirs);

	private:
		
		/// prints out the help file
		void printHelp();

		/// prints out the version info
		void printVersion();
		
		/// Finds all duplicates in path given to it
		int recursiveSearch(const fs::path &inDirPath);

		/// Compares option passed against library to see if 
		/// it is a proper parameter
		bool isFlag(string inStringToTest, int inTypeOfParam);

		/// gets the input file from the command line
		void getInput(int & inOutCurrentCounter, int inArgc, 
						  char* inArgv[]);

		/// gets the output file from the command line
		void getOutput(int &inOutCurrentCounter, int inArgc, 
							char* inArgv[]);

		/// gets paths from given inputFile
		void getPathsFromFile(fs::path inInputFile);

		/// populates list of duplicate directories
		void findDupDirs();	

		/**
       * @brief FileDup struct holds path and index of its directory
       * This structure is used in our file list of duplicates
       * and allows a file to know what directory it is in.
      */
      struct FileDup{
         fs::path path;                      //holds path
         int dirIndex;                       //Holds index of its directory

         FileDup(fs::path p1, int d1 = -1)   //constructor
         {
            path = p1;
            dirIndex = d1;
         }
      };


		/**
		 * @brief File struct holds path and vector of dups
		 * The File struct holds the path corresponding to that file
		 * as well as a vector of all duplicate files of that file
		*/
		struct File{
			fs::path path;								//holds path
			vector<FileDup> duplicates;			//holds paths of dups
			string hash;								//holds hash of file
			int dirIndex;								//Holds index of its directory		
	
			File(fs::path p1, string h1 = "", int d1 = -1)	//constructor
			{
				path = p1; 
				hash = h1;
				dirIndex = d1;
			}
		};	

	
		/**
		 * @brief CmdArguments struct holds arguments that aren't switches
		 * The CmdArguments struct holds the parameters (which aren't switches)
		 * passed by the user. It also contains a list of all the paths 
		 * given by the user.
		*/
		struct CmdArguments{
			fs::path outputFile;					//path for outputFile
			vector<fs::path> paths;				//all paths from cmd line
		};

		 /**
       * @brief Directory struct holds directory information.
       * Directory keeps track of a directory's path, the files in it (and 
		 * in its subDirs), and its sub-directories.
      */
      struct Directory{
         fs::path dirPath;             	//Path of directory
         vector<int> fGroups;		         //all the indices of child files
			vector<int> subDirs;					//all the subdirectories in folder
			vector<int> dupDirs;					//All this directory's duplicate dirs
			vector<int> mRounds;					//Bubbles down who matched who
			string hash;							//Stores hash of current files in dir
			int totalSubDirs;						//Keeps track of total subdir number
			int parent;								//Stores parent's index
			bool foundMyDups;						//If my dups already found, mark me

			//Member functions
			Directory(fs::path p1)   			//constructor
         {
            dirPath = p1;
				totalSubDirs = 0;
				hash = "";
				parent = -1;		
				foundMyDups = false;
         }
      };
		
		vector<File> fd_files;			//all the files we find
		vector<Directory> fd_dirs;		//all the directories we find
		CmdArguments fd_cmdArgs;		//our CmndArgs
		int fd_arguments;					//Contains the args that are switched

}; // end of FD class
		
//************************************************************************
#endif

